/**
 * @file
 *
 * Benchmark NUMA systems.
 *
 * @author Jens Teubner <jens.teubner@inf.ethz.ch>
 *
 * (c) 2010 ETH Zurich, Systems Group
 *
 * $Id: numabench.c 1100 2011-08-17 14:32:58Z jteubner $
 */

#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <inttypes.h>
#include <string.h>

/* FIXME: Check for getopt_long */
#include <getopt.h>

#include "setup/setup.h"

#include "workset/workset.h"
#include "pattern/pattern.h"
#include "worker/worker.h"
#include "mem/mem.h"

static void
setup_worksets (setup_t *setup)
{
    memory_regions_t regions = setup->regions;

    for (unsigned int i = 0; i < regions.count; i++)
    {
        memory_region_t *region = regions.regions[i];

        /* de-allocate memory if there was still a working set allocated */
        if (region->workset)
            nb_free (region->workset, region->ws_size);

        /* create a new working set */
        region->workset
            = generate_workset (*region->size, region->numa_nodes_mask);
        region->ws_size = *region->size;

        switch (region->pattern)
        {
            case pat_sequential:
                pattern_seq (region->workset, *region->size, *region->stride);
                break;

            case pat_random:
                pattern_rand (region->workset, *region->size, *region->stride);
                break;

            default:
                fprintf (stderr, "Illegal access pattern value.\n");
                exit (EXIT_FAILURE);
        }
    }
}

static void
prepare_threads (setup_t *setup)
{
    threads_t threads = setup->threads;

    for (unsigned int i = 0; i < threads.count; i++)
    {
        thread_t *thread = threads.threads[i];

        thread->num_runs = 5;

        if (! thread->nsec_per_step)
        {
            thread->nsec_per_step
                = malloc (thread->num_runs * sizeof (*thread->nsec_per_step));
            assert (thread->nsec_per_step);
        }
    }
}

static void
run_threads (setup_t *setup)
{
    threads_t threads = setup->threads;

    /* spawn all the threads */
    for (unsigned int i = 0; i < threads.count; i++)
    {
        thread_t *thread = threads.threads[i];

        if (pthread_create (&thread->thread, NULL, scan_worker, thread))
        {
            fprintf (stderr, "error spawning thread.\n");
            exit (EXIT_FAILURE);
        }
    }

    /* and wait for their completion */
    for (unsigned int i = 0; i < threads.count; i++)
    {
        thread_t *thread = threads.threads[i];

        pthread_join (thread->thread, NULL);
    }
}

static void
print_results (setup_t *setup, FILE *f)
{
    fprintf (f, "<run-result>\n\n");

    for (unsigned int i = 0; i < setup->parameters.count; i++)
        fprintf (f, "  <parameter name='%s' value='%lu' />\n",
                setup->parameters.params[i]->name,
                setup->parameters.params[i]->curr_value);

    fprintf (f, "\n");

    for (unsigned int i = 0; i < setup->threads.count; i++)
    {
        thread_t *thread = setup->threads.threads[i];

        fprintf (f, "  <thread id='%s'>\n", thread->id);

        for (unsigned int j = 0; j < thread->num_runs; j++)
            fprintf (f, "    <nsec-per-step run='%u'>%10.5f</nsec-per-step>\n",
                    j, thread->nsec_per_step[j]);

        fprintf (f, "  </thread>\n");
    }

    fprintf (f, "\n</run-result>\n");
}

static void
run_experiment (setup_t *setup)
{
    /* printf ("Running experiment..."); */

    setup_worksets (setup);
    prepare_threads (setup);
    run_threads (setup);

    print_results (setup, stdout);
}

static void
iterate_params (setup_t *setup, unsigned int paramnum)
{
    if (paramnum == setup->parameters.count)
        run_experiment (setup);
    else
    {
        parameter_t *param = setup->parameters.params[paramnum];

        for (unsigned int i = 0; i < param->values.count; i++)
        {
            param->curr_value = param->values.values[i];

            /* printf ("setting %s to %lu\n", param->name, param->curr_value); */

            iterate_params (setup, paramnum + 1);
        }
    }
}

int
main (int argc, char **argv)
{
    setup_t setup = (setup_t) {
        .regions = (memory_regions_t) { .count = 0, .regions = NULL }
    };

    assert (argc == 2);

    parse_setup_file (&setup, argv[1]);

    printf ("<?xml version='1.0'?>\n\n");
    printf ("<numabench-result>\n\n");

    iterate_params (&setup, 0);

    printf ("</numabench-result>\n");

    return EXIT_SUCCESS;
}

#if 0

const unsigned int sizes[] =
  { 1024, 2048, 4096, 8192, 16384, 32768, 655536, 131072, 262144,
      524288, 1048576, 2097152, 4194304, 8388608 };

/* command line options that we support */
static const struct option longopts[] = {
    { "size"       , required_argument, NULL, 'S' },
    { "help"       ,    no_argument   , NULL, 'h' },
    { "stride"     , required_argument, NULL, 's' },
    { "num-threads", required_argument, NULL, 't' }
};
static const char shortopts[] = "S:hs:t:";

static void usage (void)
{
    printf ("\n");
    printf (PACKAGE_STRING " - characterize memory performance on NUMA systems\n");
    printf ("Brought to you by Jens Teubner <jens.teubner@inf.ethz.ch>\n");
    printf ("(c) 2010 ETH Zurich, Systems Group.  All Rights Reserved.\n");
    printf ("\n");
    printf ("Available command line options:\n");
    printf ("\n");
    printf ("  -S MIN-MAX, --size=MIN-MAX\n");
    printf ("       Use memory regions between MIN and MAX bytes to scan.\n");
    printf ("       You'll want MIN to be smaller than your smallest cache\n");
    printf ("       (the default of 1024 is probably a good value).  MAX\n");
    printf ("       should be larger than the size of your largest cache,\n");
    printf ("       but not so larget that we run into out-of-memory\n");
    printf ("       problems (default: 512MB).\n");
    printf ("       Note that these are values per thread.  If you want to\n");
    printf ("       use 8 threads, you'll need more than 4GB of memory with\n");
    printf ("       the default MAX value.\n");
    printf ("\n");
    printf ("  -s MIN-MAX, --stride=MIN-MAX\n");
    printf ("       When scanning memory, access memory addresses that are\n");
    printf ("       between MIN and MAX bytes apart.  Both values must be\n");
    printf ("       multiples of %lu and greater than 0.\n", sizeof (elem_t));
    printf ("\n");
    printf ("  -t THREADS, --num-threads=THREADS\n");
    printf ("       Invoke THREADS scan threads in parallel.  This may\n");
    printf ("       stress your memory controllers and/or your core-to-core\n");
    printf ("       interconnect stronger than single-thread scans.  The\n");
    printf ("       default is to use a single thread only.\n");
    printf ("\n");
    printf ("  -h, --help\n");
    printf ("       This command line help.\n");
    printf ("\n");
    printf ("Enjoy.\n");
}

struct minmax_t {
    unsigned int min;
    unsigned int max;
};

static struct minmax_t
parse_minmax (const char *s)
{
    long           min;
    long           max;
    char          *endptr;

    min = strtol (s, &endptr, 10);

    if (min < 0 || *endptr == '\0' || ! (isspace (*endptr) || *endptr == '-'))
    {
        fprintf (stderr, "Invalid MIN-MAX argument: `%s'\n", s);
        exit (EXIT_FAILURE);
    }

    max = strtol (endptr, &endptr, 10);

    if (max > 0 || min > -max || *endptr != '\0')
    {
        fprintf (stderr, "Invalid MIN-MAX argument: `%s'\n", s);
        exit (EXIT_FAILURE);
    }

    return (struct minmax_t) { .min = min, .max = -max };
}

int
main (int argc, char **argv)
{
    char            c;
    struct minmax_t size   = { .min = 1024, .max = 512 * 1024 * 1024 };
    struct minmax_t stride = { .min = sizeof(elem_t), .max = 16 * 1024 };

    while ((c = getopt_long (argc, argv, shortopts, longopts, NULL)) != -1)
    {
        switch (c)
        {
            case 'h':                             /* --help */
                usage ();
                exit (EXIT_SUCCESS);
                break;

            case 'S':                             /* --size=MIN-MAX */
                size = parse_minmax (optarg);
                if (size.min < 4 * sizeof (elem_t))
                {
                    fprintf (stderr, "Minimum size value %u too small, "
                            "setting it to %lu.\n", size.min,
                            4 * sizeof (elem_t));
                    size.min = 4 * sizeof (elem_t);
                }
                break;

            case 's':                             /* --stride=MIN-MAX */
                stride = parse_minmax (optarg);
                if (stride.min < sizeof (elem_t))
                {
                    fprintf (stderr, "Minimum stride value %u too small, "
                            "setting it to %lu.\n", stride.min,
                            sizeof (elem_t));
                    stride.min = sizeof (elem_t);
                }
                if (stride.min % sizeof (elem_t))
                {
                    fprintf (stderr, "Minimum stride value %u not a multiple "
                            "of %lu, setting it to %lu.\n", stride.min,
                            sizeof (elem_t),
                            (stride.min / sizeof(elem_t)) * sizeof (elem_t));
                    stride.min = (stride.min/sizeof(elem_t)) * sizeof (elem_t);
                }
                if (stride.max % sizeof (elem_t))
                {
                    fprintf (stderr, "Maximum stride value %u not a multiple "
                            "of %lu, setting it to %lu.\n", stride.max,
                            sizeof (elem_t),
                            (stride.max / sizeof(elem_t)) * sizeof (elem_t));
                    stride.max = (stride.max/sizeof(elem_t)) * sizeof (elem_t);
                }
                break;

        }
    }

    for (unsigned int i = 0; i < (sizeof (sizes) / sizeof (*sizes)); i++)
    {
        elem_t       *workset;
        worker_ctx_t *w_ctx = malloc (sizeof (*w_ctx));

        assert (w_ctx);

        workset = pattern_seq (generate_workset (sizes[i]), sizes[i], 64);
        w_ctx->workset  = workset;
        w_ctx->num_runs = 3;
        w_ctx->nsec_per_step
            = malloc (w_ctx->num_runs * sizeof (*w_ctx->nsec_per_step));

        assert (w_ctx->nsec_per_step);

        /* spawn the thread */
        pthread_create (&w_ctx->thread, NULL, scan_worker, w_ctx);

        /* wait for thread to finish */
        pthread_join (w_ctx->thread, NULL);

        /* print out results */
        for (unsigned int j = 0; j < w_ctx->num_runs; j++)
            printf ("%10u  %8.4f\n", sizes[i], w_ctx->nsec_per_step[j]);
    }

    return EXIT_SUCCESS;
}
#endif
