/**
 * @file
 *
 * Generate a working set with a sequential access pattern.
 *
 * @author Jens Teubner <jens.teubner@inf.ethz.ch>
 *
 * (c) 2010 ETH Zurich, Systems Group
 *
 * $Id: sequential.c 1100 2011-08-17 14:32:58Z jteubner $
 */

#include "config.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "pattern/pattern.h"

/**
 * Fill the working set @a elems such that scanning it will lead to
 * a sequential memory access pattern (each element points to its
 * follower; the last element points back to the first).
 *
 * @param elems  working set allocated using generate_workset();
 *               starting from @a elems[0], @c next pointers will be
 *               filled by this function
 * @param size   size of array (in bytes) specified by @a elems
 * @param stride stride size; @c next pointer will point @a stride
 *               bytes ahead; must be a positive multiple of the
 *               machine pointer size
 *
 * @return pointer to @a elems
 */
elem_t *
pattern_seq (elem_t *elems, size_t size, unsigned int stride)
{
    unsigned int i;
    unsigned int next;

    assert (elems);

    if ((stride == 0) || (stride % sizeof (elem_t) != 0))
    {
        fprintf (stderr, "in pattern_seq(): stride size must be positive "
                "multiple of %lu\n", sizeof (elem_t));
        exit (EXIT_FAILURE);
    }

    i = 0;
    next = i + (stride / sizeof (elem_t));

    while (next < (size / sizeof (elem_t)))
    {
        elems[i].next = &(elems[next]);
        i = next;
        next = i + (stride / sizeof (elem_t));
    }

    elems[i].next = elems;

    return elems;
}
