/**
 * @file
 *
 * Data structures to describe a particular experimental setup.
 *
 * @author Jens Teubner <jens.teubner@inf.ethz.ch>
 *
 * (c) 2010 ETH Zurich, Systems Group
 *
 * $Id: parse.c 1100 2011-08-17 14:32:58Z jteubner $
 */

#include "config.h"

#include "setup/setup.h"

#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <libxml/xmlreader.h>

static intlist_t
intlist (const char *values)
{
    intlist_t   ret = (intlist_t) { .count = 0, .values = NULL };
    char       *endptr = (char *) values;

    while (*endptr)
    {
        unsigned long val;

        /* skip whitespace */
        while (*endptr && isspace (*endptr))
            endptr++;

        if (*endptr < '0' || *endptr > '9')
        {
            fprintf (stderr, "Illegal character in integer list: %s\n", values);
            exit (EXIT_FAILURE);
        }

        val = strtol (endptr, &endptr, 10);

        ret.count++;
        ret.values = realloc (ret.values, ret.count * sizeof (*ret.values));
        assert (ret.values);

        ret.values[ret.count-1] = val;
    }

    return ret;
}

static void
process_parameter_node (setup_t *setup, xmlTextReaderPtr reader)
{
    int   ret;
    const char *name;
    const char *value;

    parameter_t *param = malloc (sizeof (*param));
    assert (param);

    while ((ret = xmlTextReaderMoveToNextAttribute (reader)) == 1)
    {
        name = (const char *) xmlTextReaderConstName (reader);
        value = (const char *) xmlTextReaderConstValue (reader);

        if (! strcmp (name, "name"))
        {
            param->name = strdup (value);
            assert (param->name);
        }
        else if (! strcmp (name, "values"))
        {
            param->values = intlist (value);
        }
    }

    setup->parameters.count++;
    setup->parameters.params = realloc (setup->parameters.params,
            setup->parameters.count * sizeof (*setup->parameters.params));
    setup->parameters.params[setup->parameters.count-1] = param;

}

/**
 * Parse the string representation of a value as given in the XML
 * setup description file.  The string can either encode an integer
 * value or reference a variable name that was previously declared
 * with a &lt;parameter&gt; element.
 *
 * @param setup  a struct with setup information; must already
 *               contain information about declared parameters
 * @param value  string representation as in the XML setup file
 *
 * @return a pointer to an integer value; if the string contains a
 *         literal integer, memory will be allocated to hold it and
 *         a pointer to the allocated memory will be returned; if
 *         the string references a declared parameter, a pointer to
 *         the parameter's @c curr_value field will be returned
 */
static unsigned long *
parse_val (setup_t *setup, const char *value)
{
    unsigned long   val;
    unsigned long  *valptr;
    char           *endptr;

    val = strtol (value, &endptr, 10);

    if (*endptr == '\0')
    {
        /*
         * value was an integer; allocate memory for that integer
         * and return a pointer to it.
         */
        valptr = malloc (sizeof (*valptr));
        assert (valptr);

        *valptr = val;
    }
    else
    {
        /*
         * value was not an integer; interpret it as a variable name
         * and look it up in the parameter list
         */
        unsigned int i;

        for (i = 0; i < setup->parameters.count; i++)
            if (! strcmp (setup->parameters.params[i]->name, value))
                break;

        if (i == setup->parameters.count)
        {
            fprintf (stderr, "variable %s not found\n", value);
            exit (EXIT_FAILURE);
        }

        valptr = &setup->parameters.params[i]->curr_value;
    }

    return valptr;
}

static void
process_region_node (setup_t *setup, xmlTextReaderPtr reader)
{
    int   ret;
    const char *name;
    const char *value;

    /* this is the specification of a memory region */
    memory_region_t *region = malloc (sizeof (*region));
    assert (region);

    region->workset = NULL;

    while ((ret = xmlTextReaderMoveToNextAttribute (reader)) == 1)
    {
        name = (const char *) xmlTextReaderConstName (reader);
        value = (const char *) xmlTextReaderConstValue (reader);

        if (! strcmp (name, "id"))
        {
            region->id = strdup (value);
            assert (region->id);
        }
        else if (! strcmp (name, "numa-nodes"))
        {
            region->numa_nodes = strdup (value);
#ifdef HAVE_LIBNUMA
            region->numa_nodes_mask
                = numa_parse_nodestring (region->numa_nodes);
            if (! region->numa_nodes_mask)
            {
                fprintf (stderr, "Illegal numa-nodes value: %s\n", value);
                fprintf (stderr, "Consult numa_parse_nodestring() in the "
                        "numa(3) man page\n"
                        "for the format of the numa-nodes attribute.\n");
                exit (EXIT_FAILURE);
            }
#endif
        }
        else if (! strcmp (name, "size"))
            region->size = parse_val (setup, value);
        else if (! strcmp (name, "stride"))
            region->stride = parse_val (setup, value);
        else if (! strcmp (name, "pattern"))
        {
            if (! strcmp (value, "sequential"))
                region->pattern = pat_sequential;
            else if (! strcmp (value, "random"))
                region->pattern = pat_random;
            else
            {
                fprintf (stderr, "Illegal value for pattern attribute: %s\n",
                        value);
                exit (EXIT_FAILURE);
            }
        }
    }

    /* now add the new region to setup */
    setup->regions.count++;
    setup->regions.regions
        = realloc (setup->regions.regions,
                setup->regions.count * sizeof (*setup->regions.regions));
    assert (setup->regions.regions);

    setup->regions.regions[setup->regions.count-1] = region;
}

static void
process_thread_node (setup_t *setup, xmlTextReaderPtr reader)
{
    int   ret;
    const char *name;
    const char *value;

    /* this is the specification of a memory region */
    thread_t *thread = malloc (sizeof (*thread));
    assert (thread);

    thread->mode = mode_readonly;

    thread->num_runs = 0;
    thread->nsec_per_step = NULL;

    thread->numa_node = -1;

    while ((ret = xmlTextReaderMoveToNextAttribute (reader)) == 1)
    {
        name = (const char *) xmlTextReaderConstName (reader);
        value = (const char *) xmlTextReaderConstValue (reader);

        if (! strcmp (name, "id"))
        {
            thread->id = strdup (value);
            assert (thread->id);
        }
        else if (! strcmp (name, "numa-node"))
        {
            char *endptr;
            thread->numa_node = strtol (value, &endptr, 10);

            if (*endptr != '\0')
            {
                fprintf (stderr, "Illegal value for numa-node attribute: %s\n",
                        value);
                exit (EXIT_FAILURE);
            }
        }
        else if (! strcmp (name, "region"))
        {
            unsigned int i;

            for (i = 0; i < setup->regions.count; i++)
                if (! strcmp (value, setup->regions.regions[i]->id))
                    break;

            if (i >= setup->regions.count)
            {
                fprintf (stderr, "No memory region named %s declared.\n",
                        value);
                exit (EXIT_FAILURE);
            }
            thread->region = setup->regions.regions[i];
        }
        else if (! strcmp (name, "mode"))
        {
            if (! strcmp (value, "read-only"))
                thread->mode = mode_readonly;
            else if (! strcmp (value, "read-write"))
                thread->mode = mode_readwrite;
            else
            {
                fprintf (stderr, "Illegal thread mode '%s'.\n", value);
                exit (EXIT_FAILURE);
            }
        }
    }

    /* now add the new thread to setup */
    setup->threads.count++;
    setup->threads.threads
        = realloc (setup->threads.threads,
                setup->threads.count * sizeof (*setup->threads.threads));
    assert (setup->threads.threads);

    setup->threads.threads[setup->threads.count-1] = thread;
}

static void
process_node (setup_t *setup, xmlTextReaderPtr reader)
{
    const char *name = (const char *) xmlTextReaderConstName (reader);

    assert (name);

    if (! strcmp (name, "parameter"))
        process_parameter_node (setup, reader);
    else if (! strcmp (name, "region"))
        process_region_node (setup, reader);
    else if (! strcmp (name, "thread"))
        process_thread_node (setup, reader);

}

/**
 * Read in an XML setup file and populate @a setup with information
 * from the setup file.
 *
 * @param setup    struct that will be populated with information from
 *                 the setup file
 * @param filename name of an XML file with setup information; must
 *                 validate against the numabench.dtd DTD
 */
void
parse_setup_file (setup_t *setup, const char *filename)
{
    xmlTextReaderPtr  reader;
    int               ret;

    reader = xmlReaderForFile (filename, NULL,
            XML_PARSE_DTDATTR |   /* default DTD attributes */
            XML_PARSE_NOENT |     /* substitute entities */
            XML_PARSE_DTDVALID);  /* validate with the DTD */

    if (reader == NULL)
    {
        fprintf (stderr, "unable to open setup file %s\n", filename);
    }

    while ((ret = xmlTextReaderRead (reader)) == 1)
        process_node (setup, reader);

    if (ret != 0)
    {
        fprintf (stderr, "Document %s could not be parsed.\n", filename);
        exit (EXIT_FAILURE);
    }

    /* after parsing check validation result */
    if (xmlTextReaderIsValid (reader) != 1)
    {
        fprintf (stderr, "Document %s is not valid.\n", filename);
        exit (EXIT_FAILURE);
    }

    xmlFreeTextReader (reader);
}

