#!/usr/bin/env ruby
#
# Convert the XML output of numabench into a flat ASCII file
#
# Author: Jens Teubner <jens.teubner@inf.ethz.ch>
#
# (c) 2010 ETH Zurich, Systems Group
#
# == Synopsis
#
# xml2ascii.rb: converts XML output of numabench into an ASCII format
#               that is readable by gnuplot
#
# == Usage
#
# xml2ascii.rb [OPTIONS] FILENAME
#
# -h, --help:
#    this command line help
#
# -t NAME, --onlythread NAME:
#    output only data for the thread named NAME (as in the XML setup file)
#
# -S SIZE, --onlysize SIZE:
#    output only data for measurements that scanned SIZE bytes of memory
#
# -s STRIDE, --onlystride STRIDE:
#    output only data for a stride size of STRIDE
#
# FILENAME: XML file emitted by numabench
#
# Enjoy.

require 'rexml/document'
require 'getoptlong'
require 'rdoc/usage'

commandline = ARGV.join(" ")

opts = GetoptLong.new(
    [ '--help', '-h', GetoptLong::NO_ARGUMENT ],
    [ '--onlythread', '-t', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--onlysize', '-S', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--onlystride', '-s', GetoptLong::REQUIRED_ARGUMENT ]
)

xpath = '/descendant::thread'

opts.each do |opt, arg|

    case opt

    when '--help'
        RDoc::usage

    when '--onlythread'
        xpath = xpath + "[@id = '#{ arg }']"

    when '--onlysize'
        xpath = xpath + "[preceding-sibling::parameter[@name = 'size' and @value = '#{ arg }']]"

    when '--onlystride'
        xpath = xpath + "[preceding-sibling::parameter[@name = 'stride' and @value = '#{ arg }']]"

    end

end

if ARGV.length != 1
    puts "No XML filename given."
    RDoc::usage
end

f = File.new ARGV[0]
d = REXML::Document.new f

puts <<EOT
#
# numabench measurement results, converted using xml2ascii.rb
#
# Command line arguments: #{ commandline }
#
EOT

threads = REXML::XPath.match(d, xpath)

threads.each do |t|

    thread = REXML::XPath.first(t, '@id').to_s
    results = REXML::XPath.match(t, 'child::nsec-per-step')

    results.each do |r|

        size = REXML::XPath.first(
            r, 'ancestor::run-result/parameter[@name="size"]/@value').to_s
        stride = REXML::XPath.first(
            r, 'ancestor::run-result/parameter[@name="stride"]/@value').to_s
        run = REXML::XPath.first(r, '@run').to_s
        value = REXML::XPath.first(r, 'text()').to_s

        puts "#{ size } #{ stride } #{ run } #{ value }  # #{ thread }"

    end

end
